"""
COPYRIGHT

All contributions by L. Gueguen:
Copyright (c) 2016
All rights reserved.


LICENSE

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from maxtree import maxtree
import numpy as np

class MaxTree(object):
    '''
    compute max tree object from a 2d image array
    '''


    def __init__(self, im):
        '''
        construct the max tree for an image
        image shall be either uint8, uint16 or uint32
        '''
        self.layer_count=0
        self.featsDescription={}
        self._compute(im)
        self.nb_connected_components = self.mt.getNbCC()
        self.feats = np.zeros((self.nb_connected_components,0), np.float32)
    
    def __str__(self):
        st = "MaxTree of an image of size: "+ str(self.mt.getHeight())+"x"+str(self.mt.getWidth())+"\n"
        st+= "\t nb connected components: "+ str(self.nb_connected_components)+"\n"
        st+= "\t self.featsDescription.keys(): "+ str(self.featsDescription.keys())
        return st
    
    
        
    def _compute(self,im):
        if len(im.shape)!=2:
            raise IndexError('input to MaxTree should a 2d array, len(im.shape)!=2')
        if not(im.dtype == np.uint16):
            raise TypeError('input to MaxTree should be of type np.uint16')
        
        temp_im = np.ascontiguousarray(im)
        self.mt = maxtree.MT(temp_im)
        self.mt.compute()
    
    def parent(self,cc_idx):
        '''
        retrieve the index of the parent connected component
        input cc_idx: an int32 representing the index of a connected component
        return: an int32 representing the index of the parent connected component
        '''
        return self.mt.getParent(cc_idx)
    
    def contrast(self, cc_idx):
        '''
        retrieve the contrast of the current connected component with its the parent connected component
        input cc_idx: an int32 representing the index of a connected component
        return: a float32 representing the contrast with its parent
        '''
        return self.mt.getDiff(cc_idx)

    def _log_scaling(self, feats, log_scaling):
        '''
        feats come as 'xmin', 'ymin', 'xmax', 'ymax', 'area', 'angle', 'pca_big', 'pca_small', 'hu_1', 'hu_2','hu_3','hu_4','hu_5','hu_6', 'hu_7'
        we modify these to 'xmin', 'ymin', 'xmax', 'ymax', 'area', 'pca_big', 'pca_small', 'hu_1', 'hu_2','hu_3','hu_4','hu_5','hu_6', 'hu_7', 'lshape', 'cosangle', 'sinangle'
        '''
        lshape = np.sqrt(feats[:,7])/ (np.sqrt(feats[:,6])+log_scaling)
        cosangle = np.cos(feats[:,5])
        sinangle = np.sin(feats[:,5])
        feats = np.delete(feats, 5, 1)
        for j in range(4,feats.shape[1]):
            feats[:,j] = np.log(np.abs(feats[:,j]) + log_scaling)

        feats = np.concatenate((feats,lshape[:,None],cosangle[:,None],sinangle[:,None]), axis=1)
        return feats


    def compute_shape_attributes(self, log_scaling=1e-10):
        '''
        computes the shape attributes in the maxtree. The class object gets appended the attributes in self.feats
        input log_scaling: if None, no log scaling, other wise apply log scaling
        return: nothing
        note: attributes computed are 'xmin', 'ymin', 'xmax', 'ymax', 'area', 'pca_big', 'pca_small', 'hu_1', 'hu_2','hu_3','hu_4','hu_5','hu_6', 'hu_7', 'lshape', 'cosangle', 'sinangle'
        '''
        start_i = self.feats.shape[1]
        feats = self.mt.computeShapeAttributes_swig()
        feats = self._log_scaling(feats, log_scaling)
        self.feats = np.concatenate((self.feats,feats), axis=1)
        featsD=['xmin', 'ymin', 'xmax', 'ymax', 'area', 'pca_big', 'pca_small', 'hu_1', 'hu_2','hu_3','hu_4','hu_5','hu_6', 'hu_7', 'lshape', 'cosangle', 'sinangle']
        featsDescription = {f:i+start_i for i,f in enumerate(featsD)}
        self.featsDescription.update(featsDescription)
    
    def compute_layer_attributes(self, layers):
        '''
        computes the layer attributes. The class object gets appended the attributes in self.feats
        input layers: a numpy 2d or 3d array, of size w x h or w x h x nb_bands
        return: nothing
        note: attributes computed are 'average_xx', 'std_xx', 'min_val_xx', 'max_val_xx', where xx is self.layer_count
        '''
        s = layers.shape
        s+=(1,)
        if len(s)<3:
            raise IndexError('input to layer attributes should a 2d or 3 d array')
        if s[0]!=self.mt.getHeight() or s[1]!= self.mt.getWidth():
            raise IndexError('input to layer attributes should a 2d or 3 d array of same dimensions as input image')
        
        featsD=['average', 'std', 'min_val', 'max_val']
        if s[2]==1:
            loc_lyr = np.float32(layers).reshape((s[0],s[1]))
            start_i = self.feats.shape[1]
            feats = self.mt.computeLayerAttributes_swig(loc_lyr)
            self.feats = np.concatenate((self.feats,feats), axis=1)
            featsDescription={f+'_'+str(self.layer_count):i+start_i for i,f in enumerate(featsD)}
            self.featsDescription.update(featsDescription)
            self.layer_count+=1
        else:
            for i in range(s[2]):
                self.compute_layer_attributes(layers[:,:,i])
            
    
    def getAttributes(self,keys=[]):
        '''
        retrieve the attributes matrix depending on an attribute key or more than one key
        input keys: either a string or list of string
        return: a np.float32 array of size nbcc x nbattributes
        note: valid keys are given by self.featsDescription.keys()
        '''
        if len(keys)==0:
            return self.feats
        elif type(keys) is str and keys in self.featsDescription.keys():
            return self.getAttributes([keys]).squeeze()
        elif type(keys) is list:
            for k in keys:
                if k not in self.featsDescription.keys():
                    raise KeyError(k+" is not part of valid keys in self.featsDescription")
            tmp_idx = [self.featsDescription[f] for f in keys]
            return self.feats[:,tmp_idx]
        else:
            raise KeyError(keys+" is not part of valid keys in self.featsDescription")
        
    def filter(self, idx, scores = None):
        '''
        filter the max tree in direct mode, by stacking the connected components which are retained
        input idx: np.uint32 array of the indices of connected components which are retained
        input scores (optional): np.float32 array providing a score the previous indices. 
        return: a 2d array of the filtered image
        note: example of indices computation, idx = (self.getAttributes('area')>1000).nonzero()[0]
        '''
        if scores is None:
            idx_tmp = np.uint32(idx)
            return self.mt.filter_swig(idx_tmp)
        else:
            idx_tmp = np.uint32(idx)
            s_tmp = np.float32(scores)
            return self.mt.filter_swig(idx_tmp,s_tmp)

    def filterAttributes(self, idx, scores):
        '''
        filter the max tree in direct mode, by stacking the connected components scores passed
        input idx: np.uint32 array of the indices of connected components which are retained
        input scores: np.float32 array providing a score the previous indices.
        return: a 3d array of the attribute images, corresponding to count, avg, std, min, max
        note: example of indices computation, idx = (self.getAttributes('area')>1000).nonzero()[0]
                                              scores = self.getAttributes('area')[idx]
        '''
        idx_tmp = np.uint32(idx)
        s_tmp = np.float32(scores)
        out = self.mt.computePerPixelAttributes_swig(idx_tmp, s_tmp)
        o = out.reshape((5,out.shape[0]/5,out.shape[1])).transpose((1,2,0))
        return o
        


# swig -python -c++ -I../cppsrc -o maxtree_wrap.cpp maxtree.i
# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_maxtree')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_maxtree')
    _maxtree = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_maxtree', [dirname(__file__)])
        except ImportError:
            import _maxtree
            return _maxtree
        try:
            _mod = imp.load_module('_maxtree', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _maxtree = swig_import_helper()
    del swig_import_helper
else:
    import _maxtree
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0

class MT(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, MT, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, MT, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _maxtree.new_MT(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def readim(self, imarray, width, height):
        return _maxtree.MT_readim(self, imarray, width, height)

    def compute(self):
        return _maxtree.MT_compute(self)

    def filter(self, *args):
        return _maxtree.MT_filter(self, *args)

    def filter_swig(self, *args):
        return _maxtree.MT_filter_swig(self, *args)

    def coveringCC(self, *args):
        return _maxtree.MT_coveringCC(self, *args)

    def coveringCC_XY(self, *args):
        return _maxtree.MT_coveringCC_XY(self, *args)

    def computeShapeAttributes(self):
        return _maxtree.MT_computeShapeAttributes(self)

    def computeLayerAttributes(self, layer):
        return _maxtree.MT_computeLayerAttributes(self, layer)

    def computeShapeAttributes_swig(self):
        return _maxtree.MT_computeShapeAttributes_swig(self)

    def computeLayerAttributes_swig(self, imarray):
        return _maxtree.MT_computeLayerAttributes_swig(self, imarray)

    def computePerPixelAttributes(self, retained, features):
        return _maxtree.MT_computePerPixelAttributes(self, retained, features)

    def computePerPixelAttributes_swig(self, retained, score):
        return _maxtree.MT_computePerPixelAttributes_swig(self, retained, score)

    def _print(self):
        return _maxtree.MT__print(self)

    def getConnectivity(self):
        return _maxtree.MT_getConnectivity(self)

    def setConnectivity(self, connectivity):
        return _maxtree.MT_setConnectivity(self, connectivity)

    def getHeight(self):
        return _maxtree.MT_getHeight(self)

    def getIm(self):
        return _maxtree.MT_getIm(self)

    def getNbpixels(self):
        return _maxtree.MT_getNbpixels(self)

    def getWidth(self):
        return _maxtree.MT_getWidth(self)

    def serialize_swig(self):
        return _maxtree.MT_serialize_swig(self)

    def getParent(self, *args):
        return _maxtree.MT_getParent(self, *args)

    def getDiff(self, *args):
        return _maxtree.MT_getDiff(self, *args)

    def getNbCC(self):
        return _maxtree.MT_getNbCC(self)

    def __getstate__(self):
        args={}
        args['bytes'] = self.serialize_swig()
        return args
    def __setstate__(self, state):
        dummy = int(0)
        self.__init__(dummy, state['bytes'])

    __swig_destroy__ = _maxtree.delete_MT
    __del__ = lambda self: None
MT_swigregister = _maxtree.MT_swigregister
MT_swigregister(MT)

# This file is compatible with both classic and new-style classes.





#!/usr/bin/env python

from setuptools import setup, Extension
import numpy
from maxtree import __version__

maxtree_extension = Extension('_maxtree',
                              sources=['./maxtree/maxtree_wrap.cpp','./cppsrc/maxtree.cpp'],
                              include_dirs=['./cppsrc', numpy.get_include()],
                              extra_compile_args=['-std=c++11'],
                              extra_link_args=['-std=c++11'],
                              )

setup(name='maxtree',
      version=__version__,
      description='Max Tree algorithm',
      author='Lionel Gueguen',
      author_email='gueguenster@gmail.com',
      url='https://github.com/gueguenster/maxtree',
      ext_modules = [maxtree_extension],
      packages =['maxtree'],
      include_package_data=True,
     )
